-------------------------------------------------------------------------------------
-- | Namespace Definitions |---------------------------------------------------------
-------------------------------------------------------------------------------------

-- Retrieve the addon name and the namespace table.
local addon, npcg = ...

local functua 	= npcg.functua
local util 			= npcg.functua.util

local generator = { }
local data = { }
local tests = { }

-- generators and data do not exist yet
npcg.generators.faith	= generator
npcg.data.faith				= data
npcg.tests.faith			= tests

-- Note:
-- There does not seem to be a belief in the Elemental Lords as a whole, and they are not evil, but chaotic.

local faiths = { }

-- default :: { TraitDepencency } -> { TraitDepencency }
-- Provides defaults for trait dependencies, so others don't have to fill them in.
local function default(replacements)
  local replacements = replacements or { }
  return {
    races       = replacements[     "races"] or { "any" },
		tribes			= replacements[    "tribes"] or { "any" },
    alignments  = replacements["alignments"] or { "any" },
		tags				= replacements[      "tags"] or {       },
    weight      = 1
  }
end


-- a
--faiths[       "al'akir, the windlord"]  = { alignments = { "chaotic neutral" }, races = { "unknown??" }, tags = { "elemental" } } -- Note: Deceased? No known followers.
faiths[                     "animism"]	= default({ 
	alignments = {
		"lawful good",
		"neutral good",
		"lawful neutral",
    "true neutral" },
	races = {
		"night elf", 
		"orc",
		"pandaren",                                                                                                
		"tauren" },
})
faiths[                      "an'she"]	= default({
	alignments = {
		"lawful good",
		"neutral good",
		"lawful neutral" },
	races = {
	"tauren" },
})
faiths[            "ancestor worship"]  = default({
	alignments = {
		"lawful good",
		"neutral good",
		"lawful neutral",
		"true neutral",
		"lawful evil" },
	races = {
		"orc",
		"pandaren",
		"tauren" },
})
-- b
faiths[               "blood knights"]  = default({
	alignments = {
		"neutral good",
		"chaotic good",
		"lawful neutral",
		"lawful evil" },
	races = {
		"blood elf" },
	tags = {
    "light" }
})
-- c
faiths[    "church of the holy light"]  = default({
	alignments = {
    "lawful good",
		"lawful neutral",
		"neutral good",
		"true neutral" },
	races = {
    "dwarf",
		"gnome",
		"half elf",
		"half orc",
		"high elf",
		"human",
		"worgen" },
	tags = {
    "light" }
})
faiths[          "cult of the damned"]  = default({
	alignments = {
    "lawful evil" },
	races = {
		"blood elf",
		"draenei",
		"dwarf",
		"gnome",
		"half elf",
		"half orc",
		"high elf",
		"human",
		"night elf",
		"orc",
		"tauren",
		"troll",
		"forsaken" },
	tags = {
     "cult" }
})
faiths[        "the forgotten shadow"]	= default({
	races = {
   "forsaken" },
})
-- d
faiths[                  "demonology"]  = default({
	alignments = {
		"lawful neutral",
		"lawful evil",
		"neutral evil",
		"chaotic evil" },
	races = {
		"blood elf",
		"dwarf",
		"gnome",
		"goblin",
		"half elf",
		"half orc",
		"human",
		"orc",
		"forsaken",
		"worgen" },
})
faiths[                    "druidism"]  = default({
	alignments = {
		"lawful good",
		"neutral good",
		"true neutral" },
	races = {
    "dwarf",
		"goblin",
		"half elf",
		"high elf",
		"night elf",
		"tauren" },
})
-- e
faiths[                       "elune"]	= default({
	alignments = {
		"lawful good",
		"neutral good",
		"lawful neutral" },
	races = {
		"night elf" },
})
-- f
-- g
faiths[                    "geomancy"]  = default({
	alignments = {
		"neutral good" },
	races = {
		"pandaren" },
})
--[[faiths[              "goldrinn"]	= default({
	races = {
"night elf", "tauren", "dwarf", "goblin" }, tags = {
    "any" } }
]]
-- h
-- i
-- j
-- k
-- l
faiths[                       "light"]	= default({
	alignments = {
		"lawful good",
		"neutral good",
		"lawful neutral",
		"true neutral" },
	races = {
		"blood elf",
		"draenei",
		"dwarf",
		"gnome",
		"half elf",
		"high elf",
		"half orc",
		"human",
		"forsaken",
		"worgen" },
	tags = {
		"light" }
})
--[[faiths[            "lycanthoth"]	= default({
	, tags = {
    "any" } }	-- Twilight's Hammer uses this to replace Goldrinn
]]
-- m
--[[faiths[               "malorne"]	= default({
	races = {
"night elf", "tauren" }, tags = {
    "any" } }]]
--[[faiths[       "mystery of the makers"]	= default({
	alignments = {
"chaotic neutral",
														    					  								  "true neutral",
												    			  											    "chaotic good",
										    				  														"chaotic evil",
								    							  													"neutral evil" },
	races = {
    "dwarf",
						    								  																														"gnome" },
})]]
-- n
faiths[    "neptulon, the tidehunter"]  = default({
	alignments = {
		"chaotic neutral" },
	races = {
		"naga" }, tags = {
		"elemental" }
}) -- Note: Naga are not supported.
-- o
faiths[    "order of the silver hand"]  = default({
	alignments = {
		"lawful good",
		"lawful neutral" },
	races = {
		"blood elf",
		"draenei",
		"dwarf",
		"half elf",
		"half orc",
		"high elf",
		"human",
		"tauren" },
	tags = {
    "light" }
})
-- p
faiths[                      "profit"]	= default({
	alignments = {
		"chaotic neutral",
		"chaotic evil" },
	races = {
		"any" },
})
-- q
-- r
faiths[      "ragnaros, the firelord"]  = default({
	alignments = {
		"chaotic neutral" },
	races = {
    "dwarf" },
	tags = {
		"elemental" }
})
-- s
faiths[             "sect of dragons"]  = default({
	alignments = {
		"lawful good",
		"neutral good",
		"lawful neutral" },
	races = {
		"night elf" },
})
faiths[                   "shamanism"]  = default({
	alignments = {
		"lawful good",
		"neutral good",
		"chaotic good",
		"neutral good",
		"true neutral",
		"chaotic neutral" },
	races = {
		"draenei",
		"dwarf",
		"goblin",
		"half elf",
		"half orc",
		"orc",
		"pandaren",
		"tauren",
		"troll" },
})
-- t
faiths[       "the ancient guardians"]	= default({
	alignments = {
		"true neutral" },
	races = {
		"night elf",
		"tauren" },
})
faiths[                  "the arcane"]	= default({
	alignments = {
		"lawful good",
		"neutral good",
		"chaotic good",
		"neutral good",
		"true neutral",
		"chaotic neutral" },
	races = {
		"any" },
})
faiths[       "the august celestials"]  = default({
	races = {
		"panderen" },
})
faiths[          "the burning legion"]	= default({
	alignments = {
    "lawful evil" },
	races = {
		"blood elf",
		"goblin",
		"half elf",
		"human",
		"night elf",
		"orc",
		"forsaken" },
	tags = {
		"cult" }
})
faiths[            "the earth mother"]	= default({
	alignments = {
		"neutral good",
		"chaotic good",
		"chaotic neutral",
		"lawful evil",
		"neutral evil" },
	races = {
		"dwarf",
		"pandaren",
		"tauren" },
})
faiths[            "the echo of life"]	= default({
	alignments = {
		"chaotic neutral",
		"chaotic evil" },
	races = {
		"forsaken" },
})
faiths[              "the holy light"]	= default({
	alignments = {
		"lawful good",
		"neutral good",
		"lawful neutral",
		"true neutral" },
	races = {
		"blood elf",
		"draenei",
		"dwarf",
		"gnome",
		"half elf",
		"high elf",
		"half orc",
		"human",
		"forsaken",
		"worgen" },
	tags = {
    "light" }
})
faiths[                "the illidari"]  = default({
	alignments = {
		"lawful neutral",
		"chaotic neutral" },
	races = {
		"blood elf",
		"night elf" },
})
faiths[                   "the light"]	= default({
	alignments = {
		"lawful good",
		"neutral good",
		"lawful neutral",
		"true neutral" },
	races = {
		"blood elf",
		"draenei",
		"dwarf",
		"gnome",
		"half elf",
		"high elf",
		"half orc",
		"human",
		"forsaken",
		"worgen" },
	tags = {
    "light" }
})
-- Loa will be replaced by individual Loa.
faiths[                     "the loa"]	= default({
	races = {
    "troll" },
})
faiths[                "the old gods"]	= default({
	alignments = {
		"chaotic evil",
		"lawful evil",
		"neutral evil" },
	races = {
		"any" },
	tags = {
		"cult" }
})
faiths[          "the shadow council"]  = default({
	alignments = {
    "lawful evil" },
	races = {
		"orc" },
	tags = {
		"cult" }
})
faiths[                  "the titans"]	= default({
	alignments = {
		"neutral good",
		"chaotic good",
		"true neutral",
		"chaotic neutral",
		"lawful evil",
		"neutral evil",
		"chaotic evil" },
	races = {
		"any" },
})
faiths[                    "the void"]	= default({
	alignments = npcg.data.alignment.allAlignmentsExcept(
		"lawful good",
		"neutral good",
		"chaotic good",
		"true neutral",
		"chaotic neutral" ),
	races = {
		"blood elf",
		"dwarf",
		"gnome",
		"half elf",
		"half orc",
		"high elf",
		"human",
		"worgen",
		"forsaken" },
	tags = {
		"cult" }
})
--[[faiths[         "the wild gods"]  = default({
	, tags = {
    "any" } }]]
faiths[  "therazane, the stonemother"]  = default({
	alignments = {
		"lawful neutral" },
	races = {
		"dwarf",
		"tauren" },
	tags = {
		"elemental" }
})
faiths[           "twilight's hammer"]  = default({
	alignments = {
		"chaotic evil" },
	races = {
		"any" },
})
-- u
faiths[                   "undecided"]  = default({
	alignments = npcg.data.alignment.allAlignmentsExcept(
		"lawful good" ),
	races = {
		"any" },
})
--[[faiths[       "ursoc and ursol"]	= default({
	, tags = {
    "any" } }]]
-- v
faiths[                   "veritatem"]  = default({
	alignments = {
		"chaotic neutral",
		"lawful evil",
		"chaotic evil" },
	races = npcg.data.race.allRacesExcept(
		"pandaren",
		"tauren" ),
	tags = {
		"cult" }
})
faiths[                      "voodoo"]	= default({
	alignments = {
		"lawful neutral",
		"chaotic neutral",
		"lawful evil",
		"neutral evil",
		"chaotic evil" },
	races = {
    "troll" },
})
-- w
-- x
faiths[                    "%ximself"]	= default({
	alignments = {
		"chaotic neutral",
		"chaotic evil" },
	races = {
		"any" },
})
faiths[                 "%xir people"]	= default({
	alignments = {
		"lawful neutral",
		"chaotic good",
		"chaotic neutral" },
	races = {
		"any" },
})
-- y
-- z


-- Loa Faiths
faiths[                      "akunda"]	= default({
	races = {
		"troll", },
	tribes = {
		"zandalari", },
	alignments = {
		"lawful good",
		"neutral good", },
	tags = {
		"loa", },
})
faiths[                   "bwonsamdi"]	= default({
	races = {
		"troll", },
	tribes = {
		"darkspear",
		"zandalari", },
	alignments = {
		"lawful neutral",
		"true neutral",
		"lawful evil",
		"neutral evil", },
	tags = {
		"loa", },
})
faiths[                        "gonk"]	= default({
	races = {
		"troll", },
	tribes = {
		"darkspear",
		"zandalari", },
	alignments = {
		"lawful neutral",
		"true neutral",
		"chaotic neutral", },
	tags = {
		"loa", },
})
faiths[                        "gral"]	= default({
	races = {
		"troll", },
	tribes = {
		"zandalari", },
	alignments = {
		"lawful neutral",
		"true neutral", },
	tags = {
		"loa", },
})
faiths[                     "halazzi"]	= default({
	races = {
		"troll", },
	tribes = {
		"amani",
		"zandalari", },
	alignments = {
		"true neutral",
		"chaotic neutral",
		"neutral evil",
		"chaotic evil",
	},
	tags = {
		"loa", },
})
faiths[                        "jani"]	= default({
	races = {
		"troll", },
	tribes = {
		"zandalari", },
	alignments = {
		"chaotic good",
		"chaotic neutral",
		"neutral evil",
		"chaotic evil", },
	tags = {
		"loa", },
})
faiths[                     "jan'alai"]	= default({
	races = {
		"troll", },
	tribes = {
		"amani",
		"zandalari", },
	alignments = {
		"lawful neutral",
		"true neutral", },
	tags = {
		"loa", },
})
faiths[                       "kimbul"]	= default({
	races = {
		"troll", },
	tribes = {
		"farraki",
		"zandalari", },
	alignments = {
		"neutral good",
		"lawful good",
		"true neutral",
		"lawful neutral", },
	tags = {
		"loa", },
})
faiths[                      "krag'wa"]	= default({
	races = {
		"troll", },
	tribes = {
		"darkspear",
		"zandalari", },
	alignments = {
		"lawful good",
		"neutral good",
		"lawful neutral", },
	tags = {
		"loa", },
})
faiths[                         "paku"]	= default({
	races = {
		"troll", },
	tribes = {
		"zandalari", },
	alignments = {
		"chaotic neutral",
		"chaotic evil", },
	tags = {
		"loa", },
})
--[[faiths[                        "rezan"]	= default({
	
})]]
faiths[                   "shirvallah"]	= default({
		races = {
		"troll", },
	tribes = {
		"darkspear",
		"gurubashi",
		"zandalari", },
	alignments = {
		"chaotic neutral",
		"true neutral",
		"neutral evil", },
	tags = {
		"loa", },
})
faiths[                        "torga"]	= default({
	races = {
		"tortollan",
		"troll", },
	tribes = {
		"zandalari", },
	alignments = {
		"chaotic good",
		"neutral good",
		"lawful good",
		"chaotic neutral",
		"true neutral",
		"lawful neutral", },
	tags = {
		"loa", },
})

-- allFaithsExcept :: Faith -> ... -> { Faith }
function data.allFaithsExcept(...)
  local exceptions = {...}
  return functua.without(functua.keys(faiths), exceptions)
end

-- faithsByTag :: string -> { Faith }
function data.faithsByTag(tag)
  return functua.filterKeysByValue(
    function(x) return
      functua.contains(tag, x.tags)
    end,
    util.eitherOr(tag == "any", { }, faiths)
  )
end

-- generateRandomFaith :: (Alignment a, Race a) => a -> Faith
function generateRandomFaith(characterTemplate)
  return --fixGender(characterTemplate.getGender(),
    util.randomFrom(
      functua.filterKeysByValue(
        function(val) return
					(functua.contains(characterTemplate:getAlignment(), val.alignments) or functua.contains("any", val.alignments))
					and
          (functua.contains(characterTemplate:getRace(),           val.races) or functua.contains("any", val.races))
					and
					(functua.contains(characterTemplate:getTribe(),         val.tribes) or functua.contains("any", val.tribes))
        end,
        faiths
      )
    ) or "Error: No options found!"
  --)
end

-- End of File