-------------------------------------------------------------------------------------
-- | Namespace Definitions |---------------------------------------------------------
-------------------------------------------------------------------------------------

-- Retrieve the addon name and the namespace table.
local addon, npcg = ...

local functua 	= npcg.functua
local util 			= npcg.functua.util
local replicate = functua.replicate

local generator = { }
local data 			= { }
local tests 		= { }

-- generators and data do not exist yet
npcg.generators.race	= generator
npcg.data.race				= data
npcg.tests.race				= tests

---------------------------------------------------------------------------------------------------------------------------------------
-- | Raw Data | -----------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- Data Table, this will be loaded in externally. Do not use this directly, use the functions instead.
local defaultTable = {
	[	"blood elf"			] = {
		weight = 1,
		raceTypes = {
			"horde",
			"civilized",
			"elven",
			"magic", }, },
	[ "death knight"  ] = {
		weight = 1,
		raceTypes = {
			"horde",
			"alliance",
			"hero", }, },
	[	"draenei"				] = {
		weight = 1,
		raceTypes = {
			"alliance",
			"tribal",
			"magic",
			"civilized", }, },
	[	"dwarf"					] = {
		weight = 1,
		raceTypes = {
			"alliance",
			"civilized", }, },
	[	"gnome"					] = {
		weight = 1,
		raceTypes = {
			"alliance",
			"civilized",
			"magic", }, },
	[	"goblin"				] = {
		weight = 1,
		raceTypes = {
			"horde",
			"neutral",
			"civilized", }, },
	[	"half elf"			] = {
		weight = 1,
		raceTypes = {
			"horde",
			"alliance",
			"neutral",
			"civilized",
			"elven",
			"magic",
			"halfbreed", }, },
	[	"half orc"			] = {
		weight = 1,
		raceTypes = {
			"horde",
			"alliance",
			"neutral",
			"halfbreed", }, },		
	[	"high elf"			] = {
		weight = 1,
		raceTypes = {
			"alliance",
			"civilized",
			"elven",
			"magic", }, },
	[ "human"					] = {
		weight = 1,
		raceTypes = {
			"alliance",
			"civilized",
			"magic", }, },
	[ "illidari"			] = {
		weight = 1,
		raceTypes = {
			"horde",
			"alliance",
			"elven",
			"magic",
			"hero", }, },
	[ "night elf"			] = {
		weight = 1,
		raceTypes = {
			"alliance",
			"civilized",
			"elven" }, },
	[ "nightborne"		] = {
		weight = 1,
		raceTypes = {
			"horde",
			"civilized",
			"elven",
			"magic", }, },
	[ "orc"						] = {
		weight = 1,
		raceTypes = {
			"horde",
			"tribal", }, },
	[ "pandaren"			] = {
		weight = 1,
		raceTypes = {
			"horde",
			"alliance",
			"neutral", }, },
	[ "tauren"				] = {
		weight = 1,
		raceTypes = {
			"horde",
			"tribal", }, },
	[ "troll"					] = {
		weight = 1,
		raceTypes = {
			"horde",
			"tribal", }, },
	[ "forsaken"			] = {
		weight = 1,
		raceTypes = {
			"horde",
			"civilized", } },
	[ "void elf"			] = {
		weight = 1,
		raceTypes = {
				"alliance",
				"civilized",
				"elven",
				"magic",
				"hero", }, },
	[ "worgen"				] = {
		weight = 1,
		raceTypes = {
			"alliance",
			"civilized", }, },
	}

---------------------------------------------------------------------------------------------------------------------------------------
-- | Data Modifiers | -----------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions allow for customization of defaultTable
--[[ DEPRICATED!
local function defaultRaceTypes(race, ...)
	local types = {
		["blood elf"		] = {
			"horde",
			"civilized",
			"elven",
			"magic", },
		["death knight"	] = {
			"horde",
			"alliance",
			"hero", },
		["draenei"			] = {
			"alliance",
			"tribal",
			"magic",
			"civilized", },
		["dwarf"				] = {
			"alliance",
			"civilized", },
		["gnome"				] = {
			"alliance",
			"civilized",
			"magic", },
		["goblin"				] = {
			"horde",
			"neutral",
			"civilized", },
		["half elf"			] = {
			"horde",
			"alliance",
			"neutral",
			"civilized",
			"elven",
			"magic",
			"halfbreed", },
		["half orc"			] = {
			"horde",
			"alliance",
			"neutral",
			"halfbreed", },
		["high elf"			] = {
			"alliance",
			"civilized",
			"elven",
			"magic", },
		["human"				] = {
			"alliance",
			"civilized",
			"magic", },
		["illidari"			] = {
			"horde",
			"alliance",
			"elven",
			"magic",
			"hero", },
		["night elf"		] = {
			"alliance",
			"civilized",
			"elven" },
		["nightborne"		] = {
			"horde",
			"civilized",
			"elven",
			"magic", },
		["orc"					] = {
			"horde",
			"tribal", },
		["pandaren"			] = {
			"horde",
			"alliance",
			"neutral", },
		["tauren"				] = {
			"horde",
			"tribal", },
		["troll"				] = {
			"horde",
			"tribal", },
		["forsaken"			] = {
			"horde",
			"civilized", },
		["void elf"			] = {
				"alliance",
				"civilized",
				"elven",
				"magic",
				"hero", },
		["worgen"				] = {
			"alliance",
			"civilized", },
	}
	types = types[race] or { "neutral", "weird"	}
	
	local extraTypes = {...}
	for _,v in pairs(extraTypes) do
		table.insert(types, v)
	end
	
	return types
end
]]

-- default :: { Race, RaceConstraints } -> { Race, RaceConstraints }
-- Returns the default race distrubution template altered by entries within the given table.
local function default(replacements)
	local replacements = replacements or { }
	
	return functua.map2(
		function(raceKey, raceValue)
			return replacements[raceKey] or functua.clone(raceValue) end,
		defaultTable)
	--[[
	return {
		[	"blood elf"			] = replacements["blood elf"		] or {
			weight = 1,
			raceTypes = defaultRaceTypes("blood elf"),
		},
		[ "death knight"  ] = replacements["death knight"	] or {
			weight = 1,
			raceTypes = defaultRaceTypes("death knight"),
		},
		[	"draenei"				] = replacements["draenei"			] or {
			weight = 1,
			raceTypes = defaultRaceTypes("draenei"),
		},
		[	"dwarf"					] = replacements["dwarf"				] or {
			weight = 1,
			raceTypes = defaultRaceTypes("dwarf"),
		},
		[	"gnome"					] = replacements["gnome"				] or {
			weight = 1,
			raceTypes = defaultRaceTypes("gnome")
		},
		[	"goblin"				] = replacements["goblin"				] or {
			weight = 1,
			raceTypes = defaultRaceTypes("goblin")
		},
		[	"half elf"			] = replacements["half elf"			] or {
			weight = 1,
			raceTypes = defaultRaceTypes("half elf")
		},
		[	"half orc"			] = replacements["half orc"			] or {
			weight = 1,
			raceTypes = defaultRaceTypes("half orc")
		},		
		[	"high elf"			] = replacements["high elf"			] or {
			weight = 1,
			raceTypes = defaultRaceTypes("high elf")
		},
		[ "human"					] = replacements["human"				] or {
			weight = 1,
			raceTypes = defaultRaceTypes("human")
		},
		[ "illidari"			] = replacements["illidari"			] or {
			weight = 1,
			raceTypes = defaultRaceTypes("illidari") 
		},
		[ "night elf"			] = replacements["night elf"		] or {
			weight = 1,
			raceTypes = defaultRaceTypes("night elf")
		},
		[ "nightborne"		] = replacements["nightborne"		] or {
			weight = 1,
			raceTypes = defaultRaceTypes("nightborne")
		},
		[ "orc"						] = replacements["orc"					] or {
			weight = 1,
			raceTypes = defaultRaceTypes("orc")
		},
		[ "pandaren"			] = replacements["pandaren"			] or {
			weight = 1,
			raceTypes = defaultRaceTypes("pandaren")
		},
		[ "tauren"				] = replacements["tauren"				] or {
			weight = 1,
			raceTypes = defaultRaceTypes("tauren")
		},
		[ "troll"					] = replacements["troll"				] or {
			weight = 1,
			raceTypes = defaultRaceTypes("troll")
		},
		[ "forsaken"			] = replacements["forsaken"			] or {
			weight = 1,
			raceTypes = defaultRaceTypes("forsaken")
		},
		[ "void elf"			] = replacements["void elf"			] or {
			weight = 1,
			raceTypes = defaultRaceTypes("void elf") 
		},
		[ "worgen"				] = replacements["worgen"				] or {
			weight = 1,
			raceTypes = defaultRaceTypes("worgen")
		},
	}]]
end

-- defaultFromWeightTable :: { Race, Int } -> { Race, TribeConstraints }
-- Returns the default race distribution template altered by a weight table. The default weight is 1.
-- Note: If replacements contains anything not present in the default setting, it will not be returned.
local function defaultFromWeightTable(replacements)
	local output = default({ })
	
	for race, value in pairs(replacements) do
		if output[race] ~= nil then
			output[race].weight = value
		end
	end
	
	return output
	
	--[[
	return {
		[	"blood elf"			] = {
			weight = replacements["blood elf"    ] or 1,
			raceTypes = defaultRaceTypes("blood elf")
		},
		[ "death knight"  ] = {
			weight = replacements["death knight"	] or 1,
			raceTypes = defaultRaceTypes("death knight")
		},
		[	"draenei"				] = {
			weight = replacements["draenei"			] or 1,
			raceTypes = defaultRaceTypes("draenei")
		},
		[	"dwarf"					] = {
			weight = replacements["dwarf"				] or 1,
			raceTypes = defaultRaceTypes("dwarf")
		},
		[	"gnome"					] = {
			weight = replacements["gnome"				] or 1,
			raceTypes = defaultRaceTypes("gnome")
		},
		[	"goblin"				] = {
			weight = replacements["goblin"				] or 1,
			raceTypes = defaultRaceTypes("goblin")
		},
		[	"half elf"			] = {
			weight = replacements["half elf"			] or 1,
			raceTypes = defaultRaceTypes("half elf")
		},
		[	"high elf"			] = {
			weight = replacements["high elf"			] or 1,
			raceTypes = defaultRaceTypes("high elf")
		},
		[ "human"					] = {
			weight = replacements["human"				] or 1,
			raceTypes = defaultRaceTypes("human")
		},
		[ "illidari"			] = {
			weight = replacements["illidari"			] or 1,
			raceTypes = defaultRaceTypes("illidari")
		},
		[ "night elf"			] = {
			weight = replacements["night elf"		] or 1,
			raceTypes = defaultRaceTypes("night elf")
		},
		[ "nightborne"		] = {
			weight = replacements["nightborne"		] or 1,
			raceTypes = defaultRaceTypes("nightborne")
		},
		[ "orc"						] = {
			weight = replacements["orc"					] or 1,
			raceTypes = defaultRaceTypes("orc")
		},
		[ "pandaren"			] = {
			weight = replacements["pandaren"			] or 1,
			raceTypes = defaultRaceTypes("pandaren")
		},
		[ "tauren"				] = {
			weight = replacements["tauren"				] or 1,
			raceTypes = defaultRaceTypes("tauren")
		},
		[ "troll"					] = {
			weight = replacements["troll"				] or 1,
			raceTypes = defaultRaceTypes("troll")
		},
		[ "forsaken"			] = {
			weight = replacements["forsaken"			] or 1,
			raceTypes = defaultRaceTypes("forsaken")
		},
		[ "void elf"			] = {
			weight = replacements["void elf"			] or 1,
			raceTypes = defaultRaceTypes("void elf")
		},
		[ "worgen"				] = {
			weight = replacements["worgen"				] or 1,
			raceTypes = defaultRaceTypes("worgen")
		},
	}]]
end

-- defaultFromWeightTable :: { Race, Int } -> { Race, RaceConstraints }
-- Returns the default race distribution template altered by a weight table. The default weight is 0.
-- Note: If replacements contains anything not present in the default setting, it will not be returned.
local function fromWeightTable(replacements)
	local output = default({ })
	
	output = functua.map(
		function(v) local x = functua.clone(v); x.weight = 0; return x end, -- Must clone, v is a table.
		output)
	
	for race, value in pairs(replacements) do
		if output[race] ~= nil then
			output[race].weight = value
		end
	end
	
	return output

	--[[

	local replacements = replacements or { }
	return {
		[	"blood elf"			] = {
			weight = replacements["blood elf"    ] or 0,
			raceTypes = defaultRaceTypes("blood elf")
		},
		[ "death knight"  ] = {
			weight = replacements["death knight"	] or 0,
			raceTypes = defaultRaceTypes("death knight")
		},
		[	"draenei"				] = {
			weight = replacements["draenei"			] or 0,
			raceTypes = defaultRaceTypes("draenei")
		},
		[	"dwarf"					] = {
			weight = replacements["dwarf"				] or 0,
			raceTypes = defaultRaceTypes("dwarf")
		},
		[	"gnome"					] = {
			weight = replacements["gnome"				] or 0,
			raceTypes = defaultRaceTypes("gnome")
		},
		[	"goblin"				] = {
			weight = replacements["goblin"				] or 0,
			raceTypes = defaultRaceTypes("goblin")
		},
		[	"half elf"			] = {
			weight = replacements["half elf"			] or 0,
			raceTypes = defaultRaceTypes("half elf")
		},
		[	"high elf"			] = {
			weight = replacements["high elf"			] or 0,
			raceTypes = defaultRaceTypes("high elf")
		},
		[ "human"					] = {
			weight = replacements["human"				] or 0,
			raceTypes = defaultRaceTypes("human")
		},
		[ "illidari"			] = {
			weight = replacements["illidari"			] or 0,
			raceTypes = defaultRaceTypes("illidari")
		},
		[ "night elf"			] = {
			weight = replacements["night elf"		] or 0,
			raceTypes = defaultRaceTypes("night elf")
		},
		[ "nightborne"		] = {
			weight = replacements["nightborne"		] or 0,
			raceTypes = defaultRaceTypes("nightborne")
		},
		[ "orc"						] = {
			weight = replacements["orc"					] or 0,
			raceTypes = defaultRaceTypes("orc")
		},
		[ "pandaren"			] = {
			weight = replacements["pandaren"			] or 0,
			raceTypes = defaultRaceTypes("pandaren")
		},
		[ "tauren"				] = {
			weight = replacements["tauren"				] or 0,
			raceTypes = defaultRaceTypes("tauren")
		},
		[ "troll"					] = {
			weight = replacements["troll"				] or 0,
			raceTypes = defaultRaceTypes("troll")
		},
		[ "forsaken"			] = {
			weight = replacements["forsaken"			] or 0,
			raceTypes = defaultRaceTypes("forsaken")
		},
		[ "void elf"			] = {
			weight = replacements["void elf"			] or 0,
			raceTypes = defaultRaceTypes("void elf")
		},
		[ "worgen"				] = {
			weight = replacements["worgen"				] or 0,
			raceTypes = defaultRaceTypes("worgen")
		},
	}	
]]	
	
end

-- Adds the neutral raceType to all races in a table.
local function neutral(races)
	return functua.map(
		function(x)
			local output = functua.clone(x)
			output.raceTypes = functua.insert(output.raceTypes, "neutral")
			return output
			end,
		races)
end

---------------------------------------------------------------------------------------------------------------------------------------
-- | Active Data Templates | ----------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- The default race distribution template, this is called when there are no specifics for zones and/or subzones.
local races = default({ })

---------------------------------------------------------------------------------------------------------------------------------------
-- | Zone-based Generation | ----------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

local zonesToRaces = {

-- Horde Capitals
["Orgrimmar"      		] = defaultFromWeightTable({
	["goblin"			] = 4,
	["orc"				] = 22,
	["pandaren" 	] = 2,
	["tauren"			] = 2,
	["troll"			] = 6,
	["forsaken"		] = 9,
}),
["Silvermoon City"		] = defaultFromWeightTable({
	["blood elf"	] = 40,
	["half elf"		] = 2,
	["forsaken" 	] = 3,
	["nightborne"	] = 6,
}),
["Thunder Bluff"			] = defaultFromWeightTable({
	["tauren"			] = 40,
	["troll"			] = 5,
}),

-- Alliance Capitals
["Stormwind City"			] = defaultFromWeightTable({
	["dwarf"			] = 5,
	["half elf"		] = 3,
	["gnome"			] = 3,
	["high elf"		] = 4,
	["human"			] = 30,
	["night elf"	] = 8,
	["worgen"			] = 3,
}),
["City of Ironforge"	] = defaultFromWeightTable({
	["dwarf"			] = 20,
	["gnome"			] = 10,
}),
["The Exodar"					] = defaultFromWeightTable({
	["draenei"		] = 40,
}),

-- Alliance Starting Zones

-- Azuremyst Isle
["Azuremyst Isle"			] = defaultFromWeightTable({
	["draenei"		] = 100,
}),
["Ammen Vale"					] = defaultFromWeightTable({
	["draenei"		] = 100,
}),

-- Dun Morogh
["Dun Morogh"					] = defaultFromWeightTable({
	["dwarf"  		] = 80,
	["gnome"			] = 20,
}),
["Coldridge Valley"		] = defaultFromWeightTable({
	["dwarf"  		] = 90,
	["gnome"			] = 10,
}),
["New Tinkertown"			] = defaultFromWeightTable({
	["dwarf"  		] = 10,
	["gnome"			] = 90,
}),

-- Elwynn Forest
["Elwynn Forest"			] = defaultFromWeightTable({
	["human"  		] = 100,
}),
["Northshire Valley"	] = defaultFromWeightTable({
	["human"  		] = 100,
}),

-- Horde Starting Zones

-- Durotar
["Durotar"						] = defaultFromWeightTable({
	["forsaken"		] = 10,
	["orc"				] = 30,
	["troll"			] = 10,
	["tauren"			] = 3,
}),
["Valley of Trials"		] = defaultFromWeightTable({
	["orc"				] = 100,
}),
["Echo Isles"					] = defaultFromWeightTable({
	["troll"			] = 100,
}),

-- Mulgore
["Mulgore"						] = defaultFromWeightTable({
	["tauren"			] = 100,
}),
["Camp Narache"				] = defaultFromWeightTable({
	["tauren"			] = 100,
}),

-- Eversong Woods
["Eversong Woods"			] = defaultFromWeightTable({
	["blood elf"	] = 100,
	["forsaken" 	] = 10,
	["half elf"		] = 2,
}),
["Falconwing Square"	] = nil, -- Tavern in Falconwing
["Ruins of Silvermoon"] = nil, -- Sometimes replaces Eversong Woods.
["Sunstrider Isle"		] = defaultFromWeightTable({
	["blood elf"	] = 100,
}),

-- Neutral
["Dalaran"        		] = neutral(defaultFromWeightTable({
	["blood elf"	] = 10,
	["gnome"			] = 10,
	["half elf"		] = 5,
	["high elf"		] = 15,
	["human"			] = 25,
	["nightborne"	] = 5,
})),

-- Warfronts
["Arathi Highlands"		] = neutral(defaultFromWeightTable({
	["blood elf"	] = 5,
	["forsaken"		] = 5,
	["human"			] = 4,
	["high elf"		] = 2,
	["dwarf"			] = 4,
	["gnome"			] = 2,
	["worgen"			] = 5,
})),
["Darkshore"					] = neutral(defaultFromWeightTable({
	["night elf"	] = 20,
})),

-- Battle for Azeroth

-- Zandalar
["Dazar'alor"					] = defaultFromWeightTable({ -- Dazar'alor is only a select number of subzones.
	["troll"			] = 40, }),
["Nazmir"							] = defaultFromWeightTable({
	["troll"			] = 40, }),
["Vol'dun"						] = defaultFromWeightTable({
	["troll"			] = 40, }),
["Zuldazar"						] = defaultFromWeightTable({
	["troll"			] = 40, }),

-- Kul'tiras
["Boralus" 						] = defaultFromWeightTable({
	["human"			] = 40, }),
["Drustvar"						] = defaultFromWeightTable({
	["human"			] = 40, }),
["Stormsong Valley"		] = defaultFromWeightTable({
	["human"			] = 40, }),
["Tiragarde Sound"		] = defaultFromWeightTable({
	["human"			] = 40, }),

}

---------------------------------------------------------------------------------------------------------------------------------------
-- | Subzone-based Generation | -------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

subzonesToRaces = {
	["The Cape of Stranglethorn"		] = {
		["Booty Bay"									] = neutral(defaultFromWeightTable({
			["goblin"			] = 20,
		})),
		["The Old Port Authority" 		] = neutral(defaultFromWeightTable({
			["goblin"			] = 20,
		})), },
	["Booty Bay"										] = {
		["The Salty Sailor Tavern"		] = neutral(defaultFromWeightTable({
			["goblin"			] = 20,
		})), },
	["Eversong Woods"								] = {
		-- Ruins of Silvermoon.
		["Commons Hall"								] = nil, -- Big house in the back of the Ruins.
		["Dawning Lane"								] = nil,
		["Feth's Way"									] = nil, -- Right Road in Ruins.
		["Falconwing Square"					] = nil,
		["Ruins of Silvermoon"				] = nil,
		["Skulking Row"								] = nil, -- Left Road in Ruins.
	},
	["Mardum, the Shattered Abyss"	]	= {
		["The Fel Hammer"							] = neutral(fromWeightTable({
			["illidari"   ] = 1,
		})), },
--[[
  ["Orgrimmar"]   = {
    ["Valley of Honor"]     = functua.mergeTables(
      replicate(9, "orc"),
      replicate(3, "pandaren"),
      replicate(2, "goblin"),
      replicate(2, "tauren"),
      { "blood elf", "forsaken", "half elf", "half orc" }
    ),
  },
]]
	--[[
  ["Val'sharah"]  = {
    ["Bradensbrook"]        = functua.mergeTables(
      replicate( 1,      "blood elf"),
      replicate( 1,         "goblin"),
      replicate( 1,            "orc"),
      replicate( 3,         "tauren"),
      replicate( 2,          "troll"),
      replicate( 1,       "forsaken"),
      replicate( 1,        "draenei"),
      replicate( 1,          "dwarf"),
      replicate( 1,          "gnome"),
      replicate( 1,       "half elf"),
      replicate( 1,       "half orc"),
      replicate( 1,       "high elf"),
      replicate(20,          "human"),
      replicate( 1,       "pandaren"),
      replicate( 5,      "night elf"),
      replicate( 2,         "worgen"),
      replicate( 1,   "death knight")
    ),
    ["Marl's Stead"]        = functua.mergeTables(
      replicate( 1,      "blood elf"),
      replicate( 1,         "goblin"),
      replicate( 1,            "orc"),
      replicate( 3,         "tauren"),
      replicate( 2,          "troll"),
      replicate( 1,       "forsaken"),
      replicate( 1,        "draenei"),
      replicate( 1,          "dwarf"),
      replicate( 1,          "gnome"),
      replicate( 1,       "half elf"),
      replicate( 1,       "half orc"),
      replicate( 1,       "high elf"),
      replicate(20,          "human"),
      replicate( 1,       "pandaren"),
      replicate( 5,      "night elf"),
      replicate( 2,         "worgen"),
      replicate( 1,   "death knight")
    ),
    ["Carlisle Cottage"]    = functua.mergeTables(
      replicate( 1,      "blood elf"),
      replicate( 1,         "goblin"),
      replicate( 1,            "orc"),
      replicate( 3,         "tauren"),
      replicate( 2,          "troll"),
      replicate( 1,       "forsaken"),
      replicate( 1,        "draenei"),
      replicate( 1,          "dwarf"),
      replicate( 1,          "gnome"),
      replicate( 1,       "half elf"),
      replicate( 1,       "half orc"),
      replicate( 1,       "high elf"),
      replicate(20,          "human"),
      replicate( 1,       "pandaren"),
      replicate( 5,      "night elf"),
      replicate( 2,         "worgen"),
      replicate( 1,   "death knight")
    ),
    ["Heathrow Manor"]      = functua.mergeTables(
      replicate( 1,      "blood elf"),
      replicate( 1,         "goblin"),
      replicate( 1,            "orc"),
      replicate( 3,         "tauren"),
      replicate( 2,          "troll"),
      replicate( 1,       "forsaken"),
      replicate( 1,        "draenei"),
      replicate( 1,          "dwarf"),
      replicate( 1,          "gnome"),
      replicate( 1,       "half elf"),
      replicate( 1,       "half orc"),
      replicate( 1,       "high elf"),
      replicate(20,          "human"),
      replicate( 1,       "pandaren"),
      replicate( 5,      "night elf"),
      replicate( 2,         "worgen"),
      replicate( 1,   "death knight")
    ),
    ["Thornton's Cottage"]  = functua.mergeTables(
      replicate( 1,      "blood elf"),
      replicate( 1,         "goblin"),
      replicate( 1,            "orc"),
      replicate( 3,         "tauren"),
      replicate( 2,          "troll"),
      replicate( 1,       "forsaken"),
      replicate( 1,        "draenei"),
      replicate( 1,          "dwarf"),
      replicate( 1,          "gnome"),
      replicate( 1,       "half elf"),
      replicate( 1,       "half orc"),
      replicate( 1,       "high elf"),
      replicate(20,          "human"),
      replicate( 1,       "pandaren"),
      replicate( 5,      "night elf"),
      replicate( 2,         "worgen"),
      replicate( 1,   "death knight")

    ),
  },]]
}

---------------------------------------------------------------------------------------------------------------------------------------
-- | Functions for Generation and Retrieval | -----------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- Main generator should be CharacterTemplate -> { Trait }
-- This will later change to be CharacterTemplate -> { Trait, weight } or something similar.

-- Part 1: Get potential candidates
-- Part 2: Weed out unfeasable candidates, like those not within context or faction
-- Part 3: Select a random item

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 1 Functions: Generation | -------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- Maybe abstract this part??

local function checkAllLocations(zone, subZone)
  return 
    -- Test whether the SubZone has values.
    ((subzonesToRaces[zone] or { })[subZone])
    -- Test whether the Zone has values.
    or zonesToRaces[zone]
    -- Fallback to default.
    or races --zonesToRaces["Default"]
    -- Fallback in case some extreme breaking happened.
    or { }
end

local function checkZonesOnly(zone)
  return
    -- Test whether the Zone has values.
    zonesToRaces[zone]
    -- Fallback to default.
    or races --zonesToRaces["Default"]
    -- Fallback in case some extreme breaking happened.
    or { }
end

local function dontCheckLocations()
  return
    -- Fallback to default.
    races -- zonesToRaces["Default"]
    -- Fallback in case some extreme breaking happened.
    or { }
end

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 2 Functions: Filtering | --------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions should be of type:
-- Race -> bool

local function isRaceOfFaction(race)
	return functua.any(
		function(x) return x == string.lower(UnitFactionGroup("player")) or x == "neutral" end, 
		race.raceTypes)
end

-- This looks like it will backfire massively...
local function isRaceWithinContext(race, zone, subZone)
  return isRaceOfFaction(race)
	--[[
	-- REDO THIS!
  if      zone == "Suramar" then
    return isRaceOfFaction(race) or race == "night elf"
  elseif  zone == "Dalaran" then
    return true
  elseif  zone == "Broken Shore" then
    return true
  elseif  zone == "Trueshot Lodge" then
    return true
  elseif zone == "Val'sharah" and (subZone == "Bradensbrook" or subZone == "Marl's Stead" or subZone == "Carlisle Cottage" or subZone == "Heathrow Manor" or subZone == "Thornton's Cottage") then
    return isRaceOfFaction(race) or race == "human" or race == "night elf"
  elseif zone == "Val'sharah" then
    return isRaceOfFaction(race) or race == "night elf"
  elseif zone == "Azsuna" then
    return isRaceOfFaction(race) or race == "night elf"
  else
    return isRaceOfFaction(race)
  end
	]]
end

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 3 Functions: Selection | --------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions should be of type:
-- { race } -> race
-- This might change in the future.

local function selectUsingWeightedRandom(selectionList)
  return util.weightedRandomFrom(functua.keys(selectionList), function(item) return selectionList[item].weight end)
end

---------------------------------------------------------------------------------------------------------------------------------------
-- | Queries | ------------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- allRacesExcept :: Race -> ... -> { Race }
function data.allRacesExcept(...)
  return functua.without(functua.keys(races), {...})
end

-- racesByType :: RaceType -> { Race }
function data.racesByType(raceType)
  return functua.filterKeysByValue(
    function(x) return
      functua.contains(raceType, x.raceTypes)
    end,
    util.eitherOr(raceType == "any", { }, races)
  )
end

-- racesByFaction :: Faction -> { Race }
function data.racesByFaction(faction)
  return data.racesByType(string.lower(faction))
end

-- REDO THIS
-- isRaceOfFaction :: Race --> bool
function data.isRaceOfFaction(race)
	local raceWithConstraints = (	
    -- Fallback to default.
    races --zonesToRaces["Default"]
    -- Fallback in case some extreme breaking happened.
    or { })[race]
	
	return isRaceOfFaction(raceWithConstraints or {raceTypes={"neutral"}})
end

-- REDO THIS
-- isRaceWithinContext :: Race --> Zone --> Subzone --> bool
function data.isRaceWithinContext(race, zone, subZone)
	local raceWithConstraints = (
		-- Test whether the SubZone has values.
    ((subzonesToRaces[zone] or { })[subZone])
    -- Test whether the Zone has values.
    or zonesToRaces[zone]
    -- Fallback to default.
    or races --zonesToRaces["Default"]
    -- Fallback in case some extreme breaking happened.
    or { })[race]	
	return isRaceWithinContext(raceWithConstraints or {raceTypes={"neutral"}}, zone, subZone)
end

---------------------------------------------------------------------------------------------------------------------------------------
--| Combination Function | ------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

local function createRaceGenerator(generatorF, filterF, selectorF)
  return function(characterTemplate, zone, subZone) 
    return selectorF(
			functua.filter3(
				function(race) return filterF(race, zone, subZone) end, 
				generatorF(zone, subZone))) or "Error: No options found!"
  end
end

-- Context + SubZone
local contextAndSubZoneBasedRaceGenerator   = createRaceGenerator(  checkAllLocations,  isRaceWithinContext,  selectUsingWeightedRandom)
-- Context + Zone
local contextAndZoneBasedRaceGenerator      = createRaceGenerator(  checkZonesOnly,     isRaceWithinContext,  selectUsingWeightedRandom)
-- Context + No Location
local contextAndNoZoneBasedRaceGenerator    = createRaceGenerator(  dontCheckLocations, isRaceWithinContext,  selectUsingWeightedRandom)
-- Faction + SubZone
local factionAndSubZoneBasedRaceGenerator   = createRaceGenerator(  checkAllLocations,  isRaceOfFaction,      selectUsingWeightedRandom)
-- Faction + Zone
local factionAndZoneBasedRaceGenerator      = createRaceGenerator(  checkZonesOnly,     isRaceOfFaction,      selectUsingWeightedRandom)
-- Faction + No Location
local factionAndNoZoneBasedRaceGenerator    = createRaceGenerator(  dontCheckLocations, isRaceOfFaction,      selectUsingWeightedRandom)
-- No Faction + SubZone
local noFactionAndSubZoneBasedRaceGenerator = createRaceGenerator(  checkAllLocations,  functua.const(true),  selectUsingWeightedRandom)
-- No Faction + Zone
local noFactionAndZoneBasedRaceGenerator    = createRaceGenerator(  checkZonesOnly,     functua.const(true),  selectUsingWeightedRandom)
-- No Faction + No Location
local noFactionAndNoZoneBasedRaceGenerator  = createRaceGenerator(  dontCheckLocations, functua.const(true),  selectUsingWeightedRandom)

-- This function attempts to set the race generator based on settings.
function generator.initialize()
  generateRandomRace = function() return "Error: generateRandomRace is not set" end
  if not NPCG_Options.factionSettings.matching then
    -- faction matching is off
    if not NPCG_Options.locationSettings.zones then
      -- location matching is also off
      generateRandomRace = noFactionAndNoZoneBasedRaceGenerator
    elseif NPCG_Options.locationSettings.subZones then
      generateRandomRace = noFactionAndSubZoneBasedRaceGenerator
    else
      generateRandomRace = noFactionAndZoneBasedRaceGenerator
    end
  elseif NPCG_Options.factionSettings.context then
    -- faction is on, but context is also on
    if not NPCG_Options.locationSettings.zones then
      -- location matching is off
      generateRandomRace = contextAndNoZoneBasedRaceGenerator
    elseif NPCG_Options.locationSettings.subZones then
      generateRandomRace = contextAndSubZoneBasedRaceGenerator
    else
      generateRandomRace = contextAndZoneBasedRaceGenerator
    end
  else
    -- faction is on, context is off
    if not NPCG_Options.locationSettings.zones then
      -- location matching is off
      generateRandomRace = factionAndNoZoneBasedRaceGenerator
    elseif NPCG_Options.locationSettings.subZones then
      generateRandomRace = factionAndSubZoneBasedRaceGenerator
    else
      generateRandomRace = factionAndZoneBasedRaceGenerator
    end
  end
end

---------------------------------------------------------------------------------------------------------------------------------------
--| Testing Functions | ---------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

function tests.testZoneToRaceRatio()
  local test = { }
  local zone   =  util.eitherOr(
                    NPCG_Options.locationSettings.mock  and 
                    NPCG_Options.locationSettings.zones, 
                    NPCG_Options.locationSettings.mockZone,
                    GetZoneText())
  local subzone = util.eitherOr(
                    NPCG_Options.locationSettings.mockSub and
                    NPCG_Options.locationSettings.zones   and
                    NPCG_Options.locationSettings.subZones,
                    NPCG_Options.locationSettings.mockSubZone or "<No SubZone>",
                    GetSubZoneText() or "<No SubZone>")
  print("Testing Zone to Race Ratio for: " .. zone .. ", " .. subzone)
  functua.forEach(
    function(x) 
      local result = generateRandomRace(nil, zone, subzone)
      test[result] = (test[result] or 0) + 1 end,
    functua.replicate(10000, 1))
  util.logTableMapped(function(x) return x .. ", (" .. ((x / 10000) * 100) .. "%)" end, test)
end


-- End of File.