-------------------------------------------------------------------------------------
-- | Namespace Definitions |---------------------------------------------------------
-------------------------------------------------------------------------------------

-- Retrieve the addon name and the namespace table.
local addon, npcg = ...

local functua = npcg.functua
local util 		= npcg.functua.util

local generator = { }
local data 			= { }
local tests 		= { }

-- generators and data do not exist yet
npcg.generators.tribe	= generator
npcg.data.tribe				= data
npcg.tests.tribe			= tests

---------------------------------------------------------------------------------------------------------------------------------------
-- | Raw Data | -----------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- Data Table, this will be loaded in externally. Do not use this directly, use the functions instead.
local defaultTable = {
	[ "death knight"  ] = {
		[ "draenei"     ] = {
			weight = 1,
			tribeTypes = { "alliance"}, },
		[ "dwarf"       ] = {
			weight = 2,
			tribeTypes = { "alliance"}, },
		[ "gnome"       ] = {
			weight = 1,
			tribeTypes = { "alliance"}, },
		[ "goblin"      ] = {
			weight = 1,
			tribeTypes = { "horde"}, },
		[ "high elf"    ] = {
			weight = 4,
			tribeTypes = { "neutral"}, },
		[ "human"       ] = {
			weight = 5,
			tribeTypes = { "alliance"}, },
		[ "night elf"   ] = {
			weight = 1,
			tribeTypes = { "alliance"}, },
		[ "orc"         ] = {
			weight = 1,
			tribeTypes = { "horde"}, },
		[ "tauren"      ] = {
			weight = 1,
			tribeTypes = { "horde"}, },
		[ "troll"       ] = {
			weight = 2,
			tribeTypes = { "horde"}, },
		[ "worgen"      ] = {
			weight = 1,
			tribeTypes = { "alliance" } }, },
	[ "draenei"       ] = {
		[ ""            ] = {
			weight = 6,
			tribeTypes = { "alliance"}, },
		[ "lightforged" ] = {
			weight = 1,
			tribeTypes = { "alliance" } }, },
	[ "dwarf"         ] = {
		[ "dark iron"   ] = {
			weight = 1,
			tribeTypes = { "alliance"}, },
		[ "wildhammer"  ] = {
			weight = 1,
			tribeTypes = { "alliance"}, },
		[ "bronzebeard" ] = {
			weight = 1,
			tribeTypes = { "alliance"}, },
		[ "frostborn"   ] = {
			weight = 1,
			tribeTypes = { "alliance"}, },
		[ "stonefist"   ] = {
			weight = 1,
			tribeTypes = { "alliance"}, },
		[ "stormpike"   ] = {
			weight = 1,
			tribeTypes = { "alliance" } }, },
	[ "forsaken"      ] = {
		[ "blood elf"   ] = {
			weight = 1,
			tribeTypes = { "horde"}, },
		[ "high elf"    ] = {
			weight = 2,
			tribeTypes = { "horde"}, },
		[ "human"       ] = {
			weight = 7,
			tribeTypes = { "horde" } }, },
	[	"goblin"				] = {
		[	"bilgewater"	] = {
			weight = 8,
			tribeTypes = { "horde"}, },
		[ "steamweedle" ] = {
			weight = 1,
			tribeTypes = { "neutral" } }, },
	[ "human"					] = {
		[ "dalaran" 		] = {
			weight = 5,
			tribeTypes = { "alliance"}, },
		[ "lordaeron"		] = {
			weight = 1,
			tribeTypes = { "alliance"}, },
		[ "kul'tiras"   ] = {
			weight = 3,
			tribeTypes = { "alliance"}, },
		[ "stormwind"		] = {
			weight = 9,
			tribeTypes = { "alliance"}, },
		[ "stromgarde"  ] = {
			weight = 1,
			tribeTypes = { "alliance"}, },
		[ "theramore"		] = {
			weight = 1,
			tribeTypes = { "alliance" } }, },
	[ "illidari"			] = {
		[ "blood elf"		] = {
			weight = 1,
			tribeTypes = { "horde"}, },
		[ "night elf"		] = {
			weight = 2,
			tribeTypes = { "alliance" } }, },
	[ "night elf"     ] = {
		[ "kaldorei"    ] = {
			weight = 8,
			tribeTypes = { "alliance"}, },
		[ "highborne"   ] = {
			weight = 1,
			tribeTypes = { "alliance" } }, },
	[ "orc"           ] = {
		[ "fel"         ] = {
			weight = 4,
			tribeTypes = { "horde"}, },
		[ "brown"       ] = {
			weight = 1,
			tribeTypes = { "horde"}, },
		[ "mag'har"			] = {
			weight = 1,
			tribeTypes = { "horde" } }, },
	[ "pandaren"      ] = {
		[ "huojin"      ] = {
			weight = 1,
			tribeTypes = { "horde"}, },
		[ "tushui"      ] = {
			weight = 1,
			tribeTypes = { "alliance"}, },
		[ "independant" ] = {
			weight = 1,
			tribeTypes = { "neutral" } }, },
	[ "tauren"        ] = {
		[ "bloodhoof"   ] = {
			weight = 1,
			tribeTypes = { "horde"}, },
		[ "dawnstrider" ] = {
			weight = 1,
			tribeTypes = { "horde"}, },
		[ "grimtotem"   ] = {
			weight = 1,
			tribeTypes = { "horde"}, },
		[ "mistrunner"  ] = {
			weight = 1,
			tribeTypes = { "horde"}, },
		[ "ragetotem"   ] = {
			weight = 1,
			tribeTypes = { "horde"}, },
		[ "runetotem"   ] = {
			weight = 1,
			tribeTypes = { "horde"}, },
		[ "skychaser"   ] = {
			weight = 1,
			tribeTypes = { "horde"}, },
		[ "stonehoof"   ] = {
			weight = 1,
			tribeTypes = { "horde"}, },
		[ "thunderhorn" ] = {
			weight = 1,
			tribeTypes = { "horde"}, },
		[ "wildmane"    ] = {
			weight = 1,
			tribeTypes = { "horde"}, },
		[ "winterhoof"  ] = {
			weight = 1,
			tribeTypes = { "horde"}, },
		[ "clawhoof"    ] = {
			weight = 1,
			tribeTypes = { "horde"}, },
		[ "highmountain"] = {
			weight = 1,
			tribeTypes = { "horde" } }, },
	[ "troll"         ] = {
		[ "amani"       ] = {
			weight = 1,
			tribeTypes = { "horde" }, },
		[ "darkspear"   ] = {
			weight = 6,
			tribeTypes = { "horde" }, },
		[ "farraki"     ] = {
			weight = 1,
			tribeTypes = { "horde" }, },
		[ "gurubashi"   ] = {
			weight = 1,
			tribeTypes = { "horde" }, },
		[ "zandalari"   ] = {
			weight = 4,
			tribeTypes = { "horde" }, }, },
}

-- Contextual meaning of "Tribe" for the races. This is also the title displayed in the addon. Empty strings mean Tribe has no context for the race.
local context	= { 
  ["blood elf"    ] = "",
  ["death knight" ] = "Died as",
  ["draenei"      ] = "Type",
  ["dwarf"        ] = "Clan",
  ["gnome"        ] = "",
  ["goblin"       ] = "Cartel",
  ["half elf"     ] = "",
  ["half orc"     ] = "",
  ["high elf"     ] = "",
  ["human"        ] = "Origin",
	["illidari"			] = "Past",
  ["night elf"    ] = "Type",
  ["orc"          ] = "Clan",
  ["pandaren"     ] = "Faction",
  ["tauren"       ] = "Clan",
  ["troll"        ] = "Tribe",
  ["forsaken"     ] = "Died as",
  ["worgen"       ] = "",
	["a"						] = "Test",
}

---------------------------------------------------------------------------------------------------------------------------------------
-- | Data Modifiers | -----------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions allow for customization of defaultTable

-- default :: { Race, { Tribe, TribeConstraints } } -> { Race, { Tribe, TribeConstraints } }
-- Returns the default tribe distribution template altered by entries within the given table.
-- Note: If replacements contains anything not present in the default setting, it will not be returned.
local function default(replacements)
	local replacements = replacements or { }
	
	return functua.map2(
		function(race, tribes)	-- takes in the key and value
			return functua.map2(
				function(tribeKey, tribeValue) 
					return ((replacements[race] or { })[tribeKey]) or functua.clone(tribeValue) end,
				tribes)
		end,
		defaultTable)
end

-- defaultFromWeightTable :: { Race, { Tribe, Int } } -> { Race, { Tribe, TribeConstraints } }
-- Returns the default tribe distribution template altered by a weight table. The default weight is 1.
-- Note: If replacements contains anything not present in the default setting, it will not be returned.
local function defaultFromWeightTable(replacements)
	local output = default({ })
	
	for race,tribes in pairs(replacements) do
		for tribe, value in pairs(tribes) do
			if output[race] ~= nil and output[race][tribe] ~= nil then
				output[race][tribe].weight = value end
		end
	end
	return output
end

-- neutral :: { Race, { Tribe, TribeConstraints } } -> { Race, { Tribe, TribeConstraints } }
-- Returns a tribe distribution where all tribes are given a neutral tribeType.
local function neutral(entries)
	return functua.map(
		function(xs) return -- xs :: { Tribe, TribeConstraints }
			functua.map(
				function(x)
					local output = functua.map(functua.id, x) -- Clone x.
					output.tribeTypes = functua.insert(output.tribeTypes, "neutral")
					return output
				end,
				xs) end,
		entries)
end

---------------------------------------------------------------------------------------------------------------------------------------
-- | Active Data Templates | ----------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- The default tribe distribution template, this is called when there are no specifics for zones and/or subzones.
local tribes = default({ })

---------------------------------------------------------------------------------------------------------------------------------------
-- | Zone-based Generation | ----------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

local zonesToTribes = {

-- Alliance Capitals
["Stormwind City"			] = defaultFromWeightTable({
	["human"				] = {
		["stormwind"			] = 100, }, }),

-- Neutral
["Dalaran"        		] = neutral(defaultFromWeightTable({
	["human"				] = {
		["dalaran"  			] = 100, }, }) ),

-- Warfronts
["Arathi Highlands"		] = neutral(default({ })),
["Darkshore"					] = neutral(default({ })),

-- Battle for Azeroth

-- Kul'tiras
["Boralus" 						] = defaultFromWeightTable({
	["human"				] = {
		["kul'tiras"			] = 100, }, }),
["Drustvar"						] = defaultFromWeightTable({
	["human"				] = {
		["kul'tiras"			] = 100, }, }),
["Stormsong Valley"		] = defaultFromWeightTable({
	["human"				] = {
		["kul'tiras"			] = 100, }, }),
["Tiragarde Sound"		] = defaultFromWeightTable({
	["human"				] = {
		["kul'tiras"			] = 100, }, }),

-- Zandalar
["Dazar'alor"					] = defaultFromWeightTable({ -- Dazar'alor is only a select number of subzones.
	["troll"				] = {
		["zandalari"			] = 20, }, }),
["Nazmir"							] = defaultFromWeightTable({
	["troll"				] = {
		["zandalari"			] = 20, }, }),
["Vol'dun"						] = defaultFromWeightTable({
	["troll"				] = {
		["zandalari"			] = 20, }, }),
["Zuldazar"						] = defaultFromWeightTable({
	["troll"				] = {
		["zandalari"			] = 20, }, }),

}

---------------------------------------------------------------------------------------------------------------------------------------
-- | Subzone-based Generation | -------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

local subzonesToTribes = {
-- Order Halls

["Mardum, the Shattered Abyss"	]	= {
	["The Fel Hammer"] = neutral(default({ })), },
	
-- Booty Bay
["The Cape of Stranglethorn"		] = {
	["Booty Bay"							] = neutral(defaultFromWeightTable({
		["goblin"			] = {
			["bilgewater"		] = 1,
			["steamweedle"	] = 9, }, })),
	["The Old Port Authority" ] = neutral(defaultFromWeightTable({
		["goblin"			] = {
			["bilgewater"		] = 1,
			["steamweedle"	] = 9, }, })), },
["Booty Bay"										] = {
	["The Salty Sailor Tavern"] = neutral(defaultFromWeightTable({
		["goblin"			] = {
			["bilgewater"		] = 1,
			["steamweedle"	] = 9, }, })), },

}

---------------------------------------------------------------------------------------------------------------------------------------
--| Queries | -------------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- allRacesExcept :: Tribe -> ... -> { Tribe }
function data.allTribesExcept(...)
	return functua.without(functua.concatKeys(tribes), {...})
end

-- getTribeContext :: Race -> string
function data.getTribeContext(race)
	return context[race] or ""
end

---------------------------------------------------------------------------------------------------------------------------------------
-- | Functions for Generation and Retrieval | -----------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- Main generator should be CharacterTemplate -> { Trait }
-- This will later change to be CharacterTemplate -> { Trait, weight } or something similar.

-- Part 1: Get potential candidates
-- Part 2: Weed out unfeasable candidates, like those not within context or faction
-- Part 3: Select a random item

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 1 Functions: Generation | -------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- TribeDetails a => CharacterTemplate -> ZoneString -> SubZoneString -> { Tribe, a }

local function checkAllLocations(characterTemplate, zone, subZone)
  local race = characterTemplate.race
	local output =
	    -- Test whether the SubZone has values.
    ((subzonesToTribes[zone] or { })[subZone] or { })[race]
    -- Test whether the Zone has values.
    or (zonesToTribes[zone] or { })[race]
    -- Fallback to default.
    or tribes[race]
    -- Fallback in case some extreme breaking happened.
    or { [""] = { weight=1,tribeTypes= { "neutral", "weird" } } }
	
	
	
	return 
    -- Test whether the SubZone has values.
    ((subzonesToTribes[zone] or { })[subZone] or { })[race]
    -- Test whether the Zone has values.
    or (zonesToTribes[zone] or { })[race]
    -- Fallback to default.
    or tribes[race]
    -- Fallback in case some extreme breaking happened.
    or { [""] = { weight=1,tribeTypes= { "neutral", "weird" } } }
end

local function checkZonesOnly(characterTemplate, zone)
  local race = characterTemplate.race
  return
    -- Test whether the Zone has values.
    (zonesToTribes[zone] or { })[race]
    -- Fallback to default.
    or tribes[race]
    -- Fallback in case some extreme breaking happened.
    or { [""] = { weight=1,tribeTypes= { "neutral", "weird" } } }
end

local function dontCheckLocations(characterTemplate)
  local race = characterTemplate.race
  return
    -- Fallback to default.
    tribes[race]
    -- Fallback in case some extreme breaking happened.
    or { [""] = { weight=1,tribeTypes= { "neutral", "weird" } } }
end

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 2 Functions: Filtering | --------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions should be of type:
-- characterTemplate -> tribe -> zone -> subzone -> bool

local function isTribeOfFaction(characterTemplate, tribe)
	return functua.any(
		function(x)
			return (x == string.lower(UnitFactionGroup("player")) or x == "neutral") end, 
		tribe.tribeTypes)
end

local function isTribeWithinContext(characterTemplate, tribe, zone, subZone)
  --[[local race = characterTemplate.race
  if     zone == "Dalaran"      then
    return true
  elseif zone == "Trueshot Lodge" then
    return true
  elseif zone == "Val'sharah"   then
    return isTribeOfFaction(characterTemplate, tribe) or (race == "night elf" and tribe == "")
	elseif zone == "Mardum, the Shattered Abyss" and subZone == "The Fel Hammer" and race == "illidari" then
		return true
  else
    return isTribeOfFaction(characterTemplate, tribe)
  end]]
	
	return isTribeOfFaction(characterTemplate, tribe)
end

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 3 Functions: Selection | --------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions should be of type:
-- { tribe, stuff } -> tribe
-- This might change in the future.

local function selectUsingWeightedRandom(selectionList)
  return util.weightedRandomFrom(functua.keys(selectionList), function(trait) return selectionList[trait].weight end)
end

---------------------------------------------------------------------------------------------------------------------------------------
--| Combination Function | ------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

local function createTribeGenerator(generatorF, filterF, selectorF)
  return function(characterTemplate, zone, subZone)
    return selectorF(
      functua.filter3(
        function(trait) return filterF(characterTemplate, trait, zone, subZone) end, 
        generatorF(characterTemplate, zone, subZone))) or ""
  end
end

-- Context + SubZone
local contextAndSubZoneBasedTribeGenerator    = createTribeGenerator(   checkAllLocations,   isTribeWithinContext, selectUsingWeightedRandom)
-- Context + Zone
local contextAndZoneBasedTribeGenerator       = createTribeGenerator(   checkZonesOnly,      isTribeWithinContext, selectUsingWeightedRandom)
-- Context + No Location
local contextAndNoZoneBasedTribeGenerator     = createTribeGenerator(   dontCheckLocations,  isTribeWithinContext, selectUsingWeightedRandom)
-- Faction + SubZone
local factionAndSubZoneBasedTribeGenerator    = createTribeGenerator(   checkAllLocations,   isTribeOfFaction,     selectUsingWeightedRandom)
-- Faction + Zone
local factionAndZoneBasedTribeGenerator       = createTribeGenerator(   checkZonesOnly,      isTribeOfFaction,     selectUsingWeightedRandom)
-- Faction + No Location
local factionAndNoZoneBasedTribeGenerator     = createTribeGenerator(   dontCheckLocations,  isTribeOfFaction,     selectUsingWeightedRandom)
-- No Faction + SubZone
local noFactionAndSubZoneBasedTribeGenerator  = createTribeGenerator(   checkAllLocations,   functua.const(true),  selectUsingWeightedRandom)
-- No Faction + Zone
local noFactionAndZoneBasedTribeGenerator     = createTribeGenerator(   checkZonesOnly,      functua.const(true),  selectUsingWeightedRandom)
-- No Faction + No Location
local noFactionAndNoZoneBasedTribeGenerator   = createTribeGenerator(   dontCheckLocations,  functua.const(true),  selectUsingWeightedRandom)

-- Set the Tribe Generator based on settings.
-- Must be called before calling "generateRandomTribe".
-- Depends on NPCG_Options.
function generator.initialize()
  generateRandomTribe = function() return "Error: generateRandomTribe is not set" end
  if not NPCG_Options.factionSettings.matching then
    -- faction matching is off
    if not NPCG_Options.locationSettings.zones then
      -- location matching is also off
      generateRandomTribe = noFactionAndNoZoneBasedTribeGenerator
    elseif NPCG_Options.locationSettings.subZones then
      generateRandomTribe = noFactionAndSubZoneBasedTribeGenerator
    else
      generateRandomTribe = noFactionAndZoneBasedTribeGenerator
    end
  elseif NPCG_Options.factionSettings.context then
    -- faction is on, but context is also on
    if not NPCG_Options.locationSettings.zones then
      -- location matching is off
      generateRandomTribe = contextAndNoZoneBasedTribeGenerator
    elseif NPCG_Options.locationSettings.subZones then
      generateRandomTribe = contextAndSubZoneBasedTribeGenerator
    else
      generateRandomTribe = contextAndZoneBasedTribeGenerator
    end
  else
    -- faction is on, context is off
    if not NPCG_Options.locationSettings.zones then
      -- location matching is off
      generateRandomTribe = factionAndNoZoneBasedTribeGenerator
    elseif NPCG_Options.locationSettings.subZones then
      generateRandomTribe = factionAndSubZoneBasedTribeGenerator
    else
      generateRandomTribe = factionAndZoneBasedTribeGenerator
    end
  end
end

---------------------------------------------------------------------------------------------------------------------------------------
--| Testing Functions | ---------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------


-- End of File